;;; Package con funciones para implementacion de los 
;;; algoritmo A y A*
;;;
;;; A utilizar con el package "generic-search.lisp" (a generic
;;; framework for search) para implementar dichos algoritmos.
;;;
;;; Asignatura: Introducción a la Inteligencia Artificial
;;;             ETSII - ULL
;;;
;;; Profesor: Patricio García Báez (patricio@etsii.ull.es)
;;;
;;; Fecha: 17/5/2002
;;;
;;; Comentarios:
;;;           
;;;           Para la correcta utilización han de almacenarse
;;;           en las siguientes variables los nombres de las
;;;           funciones específicas a cada problema:
;;;           
;;;           *EVAL-FN* : 
;;;                       (Definida en GENERIC-SEARCH) Función que
;;;                       evalua el coste estimado de la solución
;;;                       mas cercana (f') que pasa por un estado 
;;;                       dado (g + h'). Si se desea implementar 
;;;                       el algoritmo A* esta función ha de ser 
;;;                       subestimada.
;;;                       Conviene que sea la función que acceda
;;;                       a un campo ya precalculado del estado.
;;;           
;;;           *COSTE-ACUM-FN* :
;;;                       Función que calcula el coste acumulado (g)
;;;                       para llegar a un estado dado desde el
;;;                       estado inicial.
;;;           
;;;           *IGUAL-EST-FN* :
;;;                       Predicado que compara si dos estados son
;;;                       iguales o no.
;;;
;;;           Ver el correspondiente help para cada una de ellas.
;;;           
;;;           Las variables anteriores son utilizadas por las dos
;;;           funciones genéricas que aquí se definen:  
;;;           PRIORITY-A-MERGE-FN y PRIORITY-A-DEQUEUE-FN
;;;           y que servirán respectivamente como valores de los 
;;;           parámetros :merge-fn y :dequeue-fn para la función
;;;           GENERIC-SEARCH definida en el package GENERIC-SEARCH.
;;;           
;;;           Tener en cuenta que en las anteriores funciones no se
;;;           ha impelmentado el tratamiento de estados visitados,
;;;           a parte de aquellos que se encuentran repetidos en 
;;;           la lista frente, siendo responsabilidad del usuario
;;;           su implementación si así lo cree necesario.
;;;
;;; Modificiaciones:
;;;           Mayo 2004 - Adaptarse a q_search en forma de paquete
;;;                     - No utilizar más pares punteados en 
;;;                       *search-queue*
;;;           Mayo 2006 - Ampliada documentación. Definición como
;;;                       package.
;;;           Mayo 2007 - Cambiado nombre de version y extensión
;;;                       a ".lisp"
;;;


;;; Definición del package
;;; 

(defpackage "A_STAR"
  (:use "COMMON-LISP")
  (:documentation "Funciones para implementacion de los algoritmos A y A*.")
  (:export "VERSION" "*COSTE-ACUM-FN*" "*IGUAL-EST-FN*" "CRIBA-IGUALES" "FUSION" "PRIORITY-A-MERGE-FN" "PRIORITY-A-DEQUEUE-FN"))

(in-package "A_STAR")


;;; Variable versión
;;;

(defconstant version "may-2007" 
  "Versión del package A_STAR.")


;;; Variables globales a definir
;;;

(defvar *coste-acum-fn* nil
  "Guarda la función que devuelve el coste acumulado (g) de un estado.")

(defvar *igual-est-fn* nil
  "Guarda la función que compara dos estados y dice si son iguales.")


;;; Funciones necesarias
;;;

(defun igual-y-peor-o-igual (est1 est2)
  "Indica si el estado EST1 es igual al estado EST2, pero con coste acumulado peor o igual."
  (and (funcall *igual-est-fn* est1 est2)
       (>= (funcall *coste-acum-fn* est1)
	   (funcall *coste-acum-fn* est2))))

(defun igual-y-peor (est1 est2)
  "Indica si el estado EST1 es igual al estado EST2, pero con coste acumulado peor."
  (and (funcall *igual-est-fn* est1 est2)
       (> (funcall *coste-acum-fn* est1)
	  (funcall *coste-acum-fn* est2))))

(defun criba-iguales (new-states)
  "Busca si hay elementos de la lista de hijos (NEW-STATES) en la lista de frente mediante la función almacenada en *igual-est-fn* y si es así criba los que tengan coste acumulado (función almacenada en *COSTE-ACUM-FN*) mayor de la lista de hijos o de la lista de frente (GENERIC-SEARCH:*SEARCH-QUEUE*)."
  ; Elimina los elementos de new-states que sean iguales a alguno
  ; de la lista frente pero con coste acumulado mayor o igual
  (dolist (el generic-search:*search-queue*)
    (setf new-states 
	  (delete-if #'(lambda (est) 
			 (igual-y-peor-o-igual est
					       el))
		     new-states 
		     :count 1)))
  ; Elimina los elementos de *search-queue* que sean iguales a alguno
  ; de la new-states pero con coste acumulado mayor
  (dolist (el new-states new-states)
    (setf generic-search:*search-queue* 
	  (delete-if #'(lambda (est) 
			 (igual-y-peor est el))
		     generic-search:*search-queue*
		     :count 1))))

(defun fusion (list1 list2)
  "Realiza la fusión ordenada segun f' (GENERIC-SEARCH:*EVAL-FN*) de la lista de hijos (NEW-STATES) en lista frente (GENERIC-SEARCH:*SEARCH-QUEUE*)."
  (cond
   ((endp list1) list2)
   ((endp list2) list1)
   ((<= (funcall generic-search:*eval-fn* (first list1))
	(funcall generic-search:*eval-fn* (first list2)))
    (setf (rest list1) (fusion (rest list1) list2))
    list1)
   (t (setf (rest list2) (fusion list1 (rest list2)))
      list2)))

(defun priority-a-merge-fn (new-states)
  "Fusiona la lista de hijos (NEW-STATES) a la lista de frente (GENERIC-SEARCH:*SEARCH-QUEUE*), manténiendola ordenada segun f' (GENERIC-SEARCH:*EVAL-FN*) y eliminando los peores estados de los duplicados."
  ; Chequea y criba si los estados hijos ya estan
  ; en la lista de frente de una u otra lista
  (setf new-states (criba-iguales new-states))
  ; Ordena los estados hijos en función de f'
  (setf new-states (sort new-states #'< :key generic-search:*eval-fn*))
  ; Realiza fusión ordenada de hijos en lista frente
  (fusion new-states generic-search:*search-queue*))

(defun priority-a-dequeue-fn ()
  "Función que extrae el primer elemento de la lista de frente (GENERIC-SEARCH:*SEARCH-QUEUE*) creada con PRIORITY-A-MERGE-FN devolviendo un estado."
  (pop generic-search:*search-queue*))


;;; *** A* Search ***
;;; Branch and Bound with Underestimates and Dynamic Programming.
#|
(generic-search initial-states goal-p children
		:merge-fn #'priority-a-merge-fn
		:dequeue-fn #'priority-a-dequeue-fn)
|#
