;;; Programa: Interfaz para la creación de sistemas de 
;;;           producción simples
;;;
;;; Autores:  Víctor Rodríguez Herreros
;;;           Patricio García Báez
;;;           (patricio AT etsii.ull.es)
;;;           Escuela Técnica Superior de Ingeniería
;;;           Informática (www.etsii.ull.es)
;;;
;;; Fecha: 29 de Septiembre de 1997
;;;
;;; Comentarios:
;;;    La finalidad de este programa es la de crear una 
;;;    herramienta en Lisp que permita la creación de 
;;;    sistemas de producción para la resolución de 
;;;    diferentes problemas. Por lo tanto, esta interfaz 
;;;    es independiente del tipo de problema y no hace 
;;;    ninguna suposición respecto al tipo de problema
;;;
;;; Historia:limpia-reglas
;;;   2 de Octubre de 1997 - Terminada la implementacion de 
;;;                          las funciones REGLA e HIJOS
;;;   Mayo 1998 (Patricio) - Modificaciones de maquillaje 
;;;   Mayo 2004 (Patricio) - Cambio para generar paquete shell 
;;;   Mayo 2006 (Patricio) - Cambio estructura de paquete,
;;;                          añadido variable versión. Añadida
;;;                          función limpia-reglas.
;;;   Mayo 2007 (Patricio) - Cambio nombre variable version
;;;                          y extensión a ".lisp"
;;;   Mayo 2009 (Patricio) - Capacidad para trabajar con
;;;                          varios sistemas de producción
;;;                          simultaneamente (usa-sistema)

;;; ****************************************************
;;; Documentación **************************************
;;; ****************************************************
;;;
;;; SHELL - Interfaz para la implementación de sistemas 
;;;         de producción.
;;; Programado originalmente con CLISP para OS/2.
;;; Adaptado para su uso genérico en ANSI COMMON LISP.
;;; 
;;; Uso de la libería:
;;;
;;; Una vez cargado el paquete en memoria (load) y 
;;; accesible a ser usado (use) se pueden utilizar las
;;; siguientes funciones y variables:
;;; 
;;; (USA-SISTEMA ID)
;;; Esta función crea el sistema de producción con
;;; identificador ID o, si ya existía, lo activa para su uso
;;; con las posteriores funciones. Los identificadores ID de
;;; los sitemas de producción han de ser átomos distintos
;;; para cada sistema de producción.
;;; 
;;; (REGLA PRECONDICION ACCION)
;;; 
;;; Esta función se utiliza para introducir reglas en el 
;;; sistema de producción. El primer parámetro es la función
;;; booleana que evalua una precondición para un estado 
;;; dado. El segundo parámetro se corresponde con la función
;;; que crea el estado correspondiente a la acción aplicada 
;;; sobre un estado cuando en éste se cumpla la precondición 
;;; evualuada por el primer parámetro de la función.
;;; Por tanto tanto las funciones PRECONDICION como ACCION 
;;; reciben como parámetros un estado y devuelven T o NIL y 
;;; el nuevo estado respectivamente.
;;; 
;;; (HIJOS PADRE)
;;; 
;;; Esta función se utiliza para obtener los nuevos estados 
;;; a los que se puede llegar mediante la aplicación de las 
;;; reglas del sistema de producción creado previamente 
;;; utilizando la función REGLA. Se retorna una lista con
;;; todos los estados a los que se ha podido evolucionar 
;;; desde el estado pasado como parámetro.
;;; 
;;; LIMPIA-REGLAS
;;; 
;;; Esta función vacía la variable *SISTEMA*, eliminando
;;; por tanto todas las reglas de la memoria.
;;;
;;; VERSION
;;;
;;; Esta variable almacena la versión del paquete cargada.
;;;
;;; 
;;; SISTEMA
;;; 
;;; Esta variable es de uso interno de la librería y no 
;;; debería ser modificada directamente por el usuario. 
;;; En ella se almacena el sistema de produccion creado 
;;; con la funcion REGLA.

;;; Definición del paquete

(defpackage "SHELL"
  (:nicknames "sh")
  (:use "COMMON-LISP")
  (:documentation "Interfaz para la creación de sistemas de producción simples.")
  (:export "VERSION" "USA-SISTEMA" "LIMPIA-REGLAS" "REGLA" "HIJOS"))
  
(in-package "SHELL")


;;; Variable version

(defconstant version "may-2009"
  "Versión del package SHELL.")


;;; La siguiente variable almacena todas las reglas que 
;;; se han creado para los sistemas de produccion.
;;; Se inicializa con el valor lista vacia = NIL

(defvar la-sistemas nil
  "Almacena lista asiciativa de los distintos sistemas de producción almacenados según un identificador")

;;; Lista con el sistema de producción que se está usando

(defvar sistema nil
  "Esta variable almacena en una lista con cabecera el identificador del sistema de producción en uso y a continuación todas las reglas que se han creado para dicho sistema de producción.")

(defun usa-sistema (id)
  "Indica el identificador ID del sistema de producción que se va a usar en posteriores funciones. De no existir ese sistema de producción inicializa uno nuevo con identificador ID. Si el sistema se crea devuelve ID y si ya estaba creado devuelve NIL"
  (let ((stma (assoc id la-sistemas)))
    (if stma                                    ; ¿Existía ya el sistema ID?
	(progn
	  (setf sistema stma)                     ; Usar sistema ID
	  nil)
      (progn                                      ; Crear nuevo sistema ID
	(setf la-sistemas (acons id nil la-sistemas))
	(setf sistema (first la-sistemas))
	id))))

(defun limpia-reglas ()
  "Función que se encarga de eliminar la reglas que han sido cargadas previamente mediante la función REGLA."
  (if (not sistema)                          ; Si no existia el sistema
      (usa-sistema nil))                         ; Crearlo
  (setf (rest sistema) nil))                 ; Limpiamos el sistema


;;; La función REGLA se encarga de añadir una nueva regla 
;;; al sistema de producción. Se pasan dos parámetros, la
;;; precondición de la misma y la acción a ejecutar en 
;;; caso de que la precondición se cumpla

(defun regla (precondicion accion)
  "Función que se encarga de añadir una nueva regla al sistema de producción (almacenado en la variable SISTEMA). Se le pasan dos parámetros, la precondición de la misma y la acción a ejecutar en caso de que la precondición se cumpla."
  ; Añade regla por delante en el sistema
  (push (list precondicion 
	      accion )
	(rest sistema)))


;;; La función HIJOS devuelve una lista con todos los 
;;; estados a los que se puede evolucionar siempre y 
;;; cuando se cumpla la precondición de la regla 
;;; correspondiente. El estado desde el que se genera
;;; esta lista de estados se pasar  como parámetro a  
;;; función

(defun hijos (padre)
  "Función que devuelve una lista con todos los estados a los que se puede evolucionar siempre y cuando se cumpla la precondición de la regla correspondiente. El estado desde el que se genera esta lista de estados se pasar  como parámetro a la función."
  ; La variable local LISTA almacenara a los estados 
  ; generados por la aplicación de las reglas de nuestro 
  ; sistema de producción
  (let ((lista nil))
    ; Para cada regla del sistema de producción ...
    (dolist (patron (rest sistema) lista)
      ; Si se cumple la precondición ...
      (if (funcall (first patron) padre)
        ; Ejecutamos la acción. Hay que añadir el 
        ; nuevo estado a la lista de hijos generados
	(push (funcall (first (rest patron)) padre) lista)))))

